import { field as getFieldAccessor } from "@visactor/vgrammar-util";

import { SemanticMark } from "./semantic-mark";

import { isArray, maxInArray, minInArray } from "@visactor/vutils";

import { PlotMakType } from "./enums";

import { GrammarMarkType } from "@visactor/vgrammar-core";

import { isContinuous } from "@visactor/vscale";

export class RectXSemanticMark extends SemanticMark {
    constructor(id) {
        super(PlotMakType.rectX, id);
    }
    setMarkType() {
        return GrammarMarkType.rect;
    }
    parseScaleByEncode(channel, option) {
        return this.parseScaleOfCommonEncode(channel, option);
    }
    setDefaultCrosshair() {
        return {
            x: {
                crosshairShape: "rect"
            }
        };
    }
    setDefaultTooltip() {
        var _a;
        const encodeX = null === (_a = this.spec.encode) || void 0 === _a ? void 0 : _a.x;
        return {
            content: isArray(encodeX) ? encodeX.map((entry => ({
                value: entry
            }))) : [ {
                value: encodeX
            } ]
        };
    }
    convertMarkEncode(encode) {
        var _a, _b, _c, _d, _e;
        const markEncoder = this.convertSimpleMarkEncode(encode), scaleXId = this.getScaleId("x"), res = {
            y: (datum, el, params) => 0,
            y1: (datum, el, params) => params.viewBox.height()
        };
        if (isArray(null === (_a = markEncoder.x) || void 0 === _a ? void 0 : _a.field)) res.x = {
            field: markEncoder.x.field[0],
            scale: markEncoder.x.scale
        }, res.x1 = {
            field: markEncoder.x.field[1],
            scale: markEncoder.x.scale
        }; else {
            const xAccessor = getFieldAccessor(markEncoder.x.field);
            res.x = (datum, el, params) => {
                const xVals = xAccessor(datum), scale = params[scaleXId];
                return isArray(xVals) ? scale.scale(xVals[0]) : scale.scale(xVals);
            }, res.x1 = (datum, el, params) => {
                var _a, _b, _c, _d, _e;
                const scale = params[scaleXId], xVals = xAccessor(datum);
                if (isArray(xVals) && xVals.length > 1) return scale.scale(xVals[1]);
                if (isContinuous(scale.type)) {
                    const domain = null !== (_a = scale.domain()) && void 0 !== _a ? _a : [], min = minInArray(domain), max = maxInArray(domain), baseValue = min > 0 ? min : max < 0 ? max : 0;
                    return scale.scale(baseValue);
                }
                return scale.scale(xVals) + (null !== (_e = null !== (_c = null === (_b = scale.bandwidth) || void 0 === _b ? void 0 : _b.call(scale)) && void 0 !== _c ? _c : null === (_d = scale.step) || void 0 === _d ? void 0 : _d.call(scale)) && void 0 !== _e ? _e : 0);
            };
        }
        return markEncoder.stroke && (res.stroke = markEncoder.stroke), markEncoder.color || markEncoder.group ? res.fill = null !== (_b = markEncoder.color) && void 0 !== _b ? _b : markEncoder.group : res.fill = null !== (_d = null === (_c = this.spec.style) || void 0 === _c ? void 0 : _c.fill) && void 0 !== _d ? _d : null === (_e = this.getPalette()) || void 0 === _e ? void 0 : _e[0], 
        res;
    }
}

RectXSemanticMark.type = PlotMakType.rectX;
//# sourceMappingURL=rect-x.js.map